/* $Id: sana2tags.h,v 1.7 1993/07/19 14:04:42 jraja Exp $
 *
 * Copyright (c) 1993 AmiTCP/IP Group, <amitcp-group@hut.fi>
 *                    Helsinki University of Technology, Finland.
 *                    All rights reserved.
 *
 * sana2tags.h --- Sana-II network interface parameters
 *
 * Created      : Sat Feb 13 06:44:17 1993 ppessi
 * Last modified: Mon Jul 19 17:04:03 1993 jraja
 *
 * $Log: sana2tags.h,v $
 * Revision 1.7  1993/07/19  14:04:42  jraja
 * Cleared up the arcnet stuff slightly.
 *
 * Revision 1.6  1993/06/04  11:16:15  jraja
 * Fixes for first public release.
 *
 * Revision 1.5  1993/05/16  21:09:43  ppessi
 * RCS version changed.
 *
 * Revision 1.4  1993/05/14  11:51:43  ppessi
 * Moved all interface and protocol dependent public data here.
 *
 * Revision 1.3  93/04/19  03:11:27  03:11:27  ppessi (Pekka Pessi)
 * Replaced sana_param.? with sana2tags.?
 * 
 * Revision 1.2  93/04/19  02:21:19  02:21:19  ppessi (Pekka Pessi)
 * Added IFS2_ARPHRD and IFS2_RESET tags.
 */

#ifndef SANA2TAGS_H
#define SANA2TAGS_H

#include <exec/types.h>
#include <utility/tagitem.h>

/*
 * parameter table for different wire types
 */
struct wiretype_parameters
{
  ULONG  wiretype;		/* from RFC */
  WORD   flags;			/* initial iff_flags */
  struct TagItem *tags;
};

/*
 * Different Parameter Tags
 */
#define IFS2_Dummy       (TAG_USER + 0xB0000)
#define IFS2_TOTAL       (IFS2_Dummy + 1)
#define IFS2_IPNO	 (IFS2_Dummy + 2)
#define IFS2_IPTYPE	 (IFS2_Dummy + 3)
#define IFS2_ARPNO	 (IFS2_Dummy + 4)
#define IFS2_ARPTYPE	 (IFS2_Dummy + 5)
#define IFS2_ARPHRD      (IFS2_Dummy + 6)
#define IFS2_ISONO       (IFS2_Dummy + 7)
#define IFS2_ISOTYPE     (IFS2_Dummy + 8)
#define IFS2_CCITTNO     (IFS2_Dummy + 9)
#define IFS2_CCITTTYPE   (IFS2_Dummy + 10)
#define IFS2_NSNO        (IFS2_Dummy + 11)
#define IFS2_NSTYPE      (IFS2_Dummy + 12)
#define IFS2_RAWNO       (IFS2_Dummy + 13)
#define IFS2_RESET       (IFS2_Dummy + 14)

/*
 * Well known parameters for IP 
 */
#define	ETHERTYPE_IP	0x0800		/* IP protocol */
#define ETHERTYPE_ARP	0x0806		/* Addr. resolution protocol */

#define SLIPTYPE_IP ETHERTYPE_IP
#define CSLIPTYPE_IP ETHERTYPE_IP

#define ARCOTYPE_IP     240	/* RFC 1051 framing */
#define ARCOTYPE_ARP    241	/* RFC 1051 framing */

#define ARCNTYPE_IP     212	/* RFC 1201 framing */
#define ARCNTYPE_ARP    213	/* RFC 1201 framing */
#define ARCNTYPE_RARP   214	/* RFC 1201 framing */

#endif
