RCS_ID_C = "$Id: sprint.c,v 6.2 93/10/15 03:02:28 ppessi Exp $";

/*
 * sprint.c - print finger information in short format
 *
 * Copyright  1993 AmiTCP/IP Group, <AmiTCP-group@hut.fi>
 *                  Helsinki University of Technology, Finland.
 *
 * This program is derived from original work distributed in BSD Net 2.
 *
 * Created      : Fri Oct 15 01:29:07 1993 ppessi
 * Last modified: Fri Oct 15 01:49:31 1993 ppessi
 *
 * $Log:	sprint.c,v $
 * Revision 6.2  93/10/15  03:02:28  ppessi
 * Removed -s support (until we got utmp from somewhere)
 * 
 */
 
/*
 * Copyright (c) 1989 The Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Tony Nardo of the Johns Hopkins University/Applied Physics Lab.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/time.h>
#include <tzfile.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "finger.h"

extern int entries;

PERSON ** sort(void);
int psort(PERSON **p, PERSON **t);
void stimeprint(WHERE *w);

void
sflag_print(void)
{
  extern time_t now;
  register PERSON *pn;
  register WHERE *w;
  register int cnt;
  register char *p;
  PERSON **list, **sort();
  time_t time();
  char *ctime(), *prphone();

  list = sort();
  /*
   * short format --
   *	login name
   *	real name
   *	terminal name (the XX of ttyXX)
   *	if terminal writeable (add an '*' to the terminal name
   *		if not)
   *	if logged in show idle time and day logged in, else
   *		show last login date and time.  If > 6 moths,
   *		show year instead of time.
   *	office location
   *	office phone
   */
#define	MAXREALNAME	20
#define MAXREALNAME_S  "20"
  (void)printf("%-*s %-*s %s\n", UT_NAMESIZE, "Login", MAXREALNAME,
	       "Name", "Tty  Idle  Login Time   Office     Office Phone");
  for (cnt = 0; cnt < entries; ++cnt) {
    pn = list[cnt];
    for (w = pn->whead; w != NULL; w = w->next) {
      (void)printf("%-8.8s "
		   "%-" MAXREALNAME_S "." MAXREALNAME_S "s ", 
		   pn->name, 
		   pn->realname ? pn->realname : "");
      if (!w->loginat) {
	(void)printf("  *     *  No logins   ");
	goto office;
      }
      (void)putchar(w->info == LOGGEDIN && !w->writable ?
		    '*' : ' ');
      if (*w->tty)
	(void)printf("%-2.2s ",
		     w->tty[0] != 't' || w->tty[1] != 't' ||
		     w->tty[2] != 'y' ? w->tty : w->tty + 3);
      else
	(void)printf("   ");
      if (w->info == LOGGEDIN) {
	stimeprint(w);
	(void)printf("  ");
      } else
	(void)printf("    *  ");
      p = ctime(&w->loginat);
      (void)printf("%.6s", p + 4);
      if (now - w->loginat >= SECSPERDAY * DAYSPERNYEAR / 2)
	(void)printf("  %.4s", p + 20);
      else
	(void)printf(" %.5s", p + 11);
    office:			if (pn->office)
      (void)printf(" %-10.10s", pn->office);
    else if (pn->officephone)
      (void)printf(" %-10.10s", " ");
      if (pn->officephone)
	(void)printf(" %-.15s",
		     prphone(pn->officephone));
      putchar('\n');
    }
  }
}

PERSON **
sort(void)
{
  register PERSON *pn, **lp;
  PERSON **list;
  int psort();

  if (!(list = (PERSON **)malloc((u_int)(entries * sizeof(PERSON *))))) {
    (void)fprintf(stderr, "finger: out of space.\n");
    exit(1);
  }
  for (lp = list, pn = phead; pn != NULL; pn = pn->next)
    *lp++ = pn;
  (void)qsort(list, entries, sizeof(PERSON *), psort);
  return(list);
}

int
psort(PERSON **p, PERSON **t)
{
  return(strcmp((*p)->name, (*t)->name));
}

void
stimeprint(WHERE *w)
{
  register struct tm *delta;

  delta = gmtime(&w->idletime);
  if (!delta->tm_yday)
    if (!delta->tm_hour)
      if (!delta->tm_min)
	(void)printf("     ");
      else
	(void)printf("%5d", delta->tm_min);
    else
      (void)printf("%2d:%02d",
		   delta->tm_hour, delta->tm_min);
  else
    (void)printf("%4dd", delta->tm_yday);
}
